@@ -2,8 +2,9 @@
  * Read a squashfs filesystem.  This is a highly compressed read only
  * filesystem.
  *
- * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
- * Phillip Lougher <phillip@lougher.demon.co.uk>
+ * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
+ * 2012, 2013, 2014
+ * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -31,6 +32,8 @@
 #include <errno.h>
 #include <string.h>
 #include <sys/mman.h>
+#include <limits.h>
+#include <dirent.h>
 
 #ifndef linux
 #define __BYTE_ORDER BYTE_ORDER
@@ -42,95 +45,96 @@
 
 #include <stdlib.h>
 
-#ifdef SQUASHFS_TRACE
-#define TRACE(s, args...) \
-		do { \
-			printf("mksquashfs: "s, ## args); \
-		} while(0)
-#else
-#define TRACE(s, args...)
-#endif
-
-#define ERROR(s, args...) \
-		do { \
-			fprintf(stderr, s, ## args); \
-		} while(0)
-
 #include "squashfs_fs.h"
 #include "squashfs_swap.h"
-#include "read_fs.h"
 #include "compressor.h"
 #include "xattr.h"
+#include "error.h"
+#include "mksquashfs.h"
 
-extern int read_fs_bytes(int, long long, int, void *);
-extern int add_file(long long, long long, long long, unsigned int *, int,
-	unsigned int, int, int);
-extern void *create_id(unsigned int);
-extern unsigned int get_uid(unsigned int);
-extern unsigned int get_guid(unsigned int);
-
-static struct compressor *comp;
-
-int read_block(int fd, long long start, long long *next, void *block)
+int read_block(int fd, long long start, long long *next, int expected,
+								void *block)
 {
 	unsigned short c_byte;
-	int res;
+	int res, compressed;
+	int outlen = expected ? expected : SQUASHFS_METADATA_SIZE;
 	
+	/* Read block size */
 	res = read_fs_bytes(fd, start, 2, &c_byte);
 	if(res == 0)
 		return 0;
 
 	SQUASHFS_INSWAP_SHORTS(&c_byte, 1);
+	compressed = SQUASHFS_COMPRESSED(c_byte);
+	c_byte = SQUASHFS_COMPRESSED_SIZE(c_byte);
 
-	if(SQUASHFS_COMPRESSED(c_byte)) {
-		char buffer[SQUASHFS_METADATA_SIZE];
-		int error, res;
+	/*
+	 * The block size should not be larger than
+	 * the uncompressed size (or max uncompressed size if
+	 * expected is 0)
+	 */
+	if (c_byte > outlen)
+		return 0;
+
+	if(compressed) {
+		char buffer[c_byte];
+		int error;
 
-		c_byte = SQUASHFS_COMPRESSED_SIZE(c_byte);
 		res = read_fs_bytes(fd, start + 2, c_byte, buffer);
 		if(res == 0)
 			return 0;
 
 		res = compressor_uncompress(comp, block, buffer, c_byte,
-			SQUASHFS_METADATA_SIZE, &error);
+			outlen, &error);
 		if(res == -1) {
 			ERROR("%s uncompress failed with error code %d\n",
 				comp->name, error);
 			return 0;
 		}
-		if(next)
-			*next = start + 2 + c_byte;
-		return res;
 	} else {
-		c_byte = SQUASHFS_COMPRESSED_SIZE(c_byte);
 		res = read_fs_bytes(fd, start + 2, c_byte, block);
 		if(res == 0)
 			return 0;
-
-		if(next)
-			*next = start + 2 + c_byte;
-		return c_byte;
+		res = c_byte;
 	}
+
+	if(next)
+		*next = start + 2 + c_byte;
+
+	/*
+	 * if expected, then check the (uncompressed) return data
+	 * is of the expected size
+	 */
+	if(expected && expected != res)
+		return 0;
+	else
+		return res;
 }
 
 
+#define NO_BYTES(SIZE) \
+	(bytes - (cur_ptr - *inode_table) < (SIZE))
+
+#define NO_INODE_BYTES(INODE) NO_BYTES(sizeof(struct INODE))
+
 int scan_inode_table(int fd, long long start, long long end,
 	long long root_inode_start, int root_inode_offset,
-	struct squashfs_super_block *sBlk, union squashfs_inode_header *dir_inode,
-	unsigned char **inode_table, unsigned int *root_inode_block,
+	struct squashfs_super_block *sBlk, union squashfs_inode_header
+	*dir_inode, unsigned char **inode_table, unsigned int *root_inode_block,
 	unsigned int *root_inode_size, long long *uncompressed_file,
 	unsigned int *uncompressed_directory, int *file_count, int *sym_count,
 	int *dev_count, int *dir_count, int *fifo_count, int *sock_count,
 	unsigned int *id_table)
 {
 	unsigned char *cur_ptr;
-	int byte, bytes = 0, size = 0, files = 0;
-	struct squashfs_reg_inode_header inode;
-	unsigned int directory_start_block;
+	int byte, files = 0;
+	unsigned int directory_start_block, bytes = 0, size = 0;
+	struct squashfs_base_inode_header base;
 
 	TRACE("scan_inode_table: start 0x%llx, end 0x%llx, root_inode_start "
 		"0x%llx\n", start, end, root_inode_start);
 
+	*root_inode_block = UINT_MAX;
 	while(start < end) {
 		if(start == root_inode_start) {
 			TRACE("scan_inode_table: read compressed block 0x%llx "
@@ -141,232 +145,349 @@ int scan_inode_table(int fd, long long start, long long end,
 			*inode_table = realloc(*inode_table, size
 				+= SQUASHFS_METADATA_SIZE);
 			if(*inode_table == NULL)
-				return FALSE;
+				MEM_ERROR();
 		}
 		TRACE("scan_inode_table: reading block 0x%llx\n", start);
-		byte = read_block(fd, start, &start, *inode_table + bytes);
-		if(byte == 0) {
-			free(*inode_table);
-			return FALSE;
-		}
+		byte = read_block(fd, start, &start, 0, *inode_table + bytes);
+		if(byte == 0)
+			goto corrupted;
+
 		bytes += byte;
+
+		/* If this is not the last metadata block in the inode table
+		 * then it should be SQUASHFS_METADATA_SIZE in size.
+		 * Note, we can't use expected in read_block() above for this
+		 * because we don't know if this is the last block until
+		 * after reading.
+		 */
+		if(start != end && byte != SQUASHFS_METADATA_SIZE)
+			goto corrupted;
 	}
 
+	/*
+	 * We expect to have found the metadata block containing the
+	 * root inode in the above inode_table metadata block scan.  If it
+	 * hasn't been found then the filesystem is corrupted
+	 */
+	if(*root_inode_block == UINT_MAX)
+		goto corrupted;
+
+	/*
+	 * The number of bytes available after the root inode medata block
+	 * should be at least the root inode offset + the size of a
+	 * regular directory inode, if not the filesystem is corrupted
+	 *
+	 *	+-----------------------+-----------------------+
+	 *	| 			|        directory	|
+	 *	|			|          inode	|
+	 *	+-----------------------+-----------------------+
+	 *	^			^			^
+	 *	*root_inode_block	root_inode_offset	bytes
+	 */
+	if((bytes - *root_inode_block) < (root_inode_offset +
+			sizeof(struct squashfs_dir_inode_header)))
+		goto corrupted;
+
 	/*
 	 * Read last inode entry which is the root directory inode, and obtain
 	 * the last directory start block index.  This is used when calculating
 	 * the total uncompressed directory size.  The directory bytes in the
 	 * last * block will be counted as normal.
 	 *
-	 * The root inode is ignored in the inode scan.  This ensures there is
-	 * always enough bytes left to read a regular file inode entry
+	 * Note, the previous check ensures the following calculation won't
+	 * underflow, and we won't access beyond the buffer
 	 */
 	*root_inode_size = bytes - (*root_inode_block + root_inode_offset);
 	bytes = *root_inode_block + root_inode_offset;
-	SQUASHFS_SWAP_BASE_INODE_HEADER(&dir_inode->base, *inode_table + bytes);
-	if(dir_inode->base.inode_type == SQUASHFS_DIR_TYPE) {
-		SQUASHFS_SWAP_DIR_INODE_HEADER(&dir_inode->dir,
-			*inode_table + bytes);
+	SQUASHFS_SWAP_DIR_INODE_HEADER(*inode_table + bytes, &dir_inode->dir);
+	
+	if(dir_inode->base.inode_type == SQUASHFS_DIR_TYPE)
 		directory_start_block = dir_inode->dir.start_block;
-	} else {
-		SQUASHFS_SWAP_LDIR_INODE_HEADER(&dir_inode->ldir,
-			*inode_table + bytes);
+	else if(dir_inode->base.inode_type == SQUASHFS_LDIR_TYPE) {
+		if(*root_inode_size < sizeof(struct squashfs_ldir_inode_header))
+			/* corrupted filesystem */
+			goto corrupted;
+		SQUASHFS_SWAP_LDIR_INODE_HEADER(*inode_table + bytes,
+			&dir_inode->ldir);
 		directory_start_block = dir_inode->ldir.start_block;
-	}
+	} else
+		/* bad type, corrupted filesystem */
+		goto corrupted;
+
 	get_uid(id_table[dir_inode->base.uid]);
 	get_guid(id_table[dir_inode->base.guid]);
 
+	/* allocate fragment to file mapping table */
+	file_mapping = calloc(sBlk->fragments, sizeof(struct append_file *));
+	if(file_mapping == NULL)
+		MEM_ERROR();
+
 	for(cur_ptr = *inode_table; cur_ptr < *inode_table + bytes; files ++) {
-		SQUASHFS_SWAP_REG_INODE_HEADER(&inode, cur_ptr);
+		if(NO_INODE_BYTES(squashfs_base_inode_header))
+			/* corrupted filesystem */
+			goto corrupted;
+
+		SQUASHFS_SWAP_BASE_INODE_HEADER(cur_ptr, &base);
 
 		TRACE("scan_inode_table: processing inode @ byte position "
 			"0x%x, type 0x%x\n",
 			(unsigned int) (cur_ptr - *inode_table),
-			inode.inode_type);
+			base.inode_type);
 
-		get_uid(id_table[inode.uid]);
-		get_guid(id_table[inode.guid]);
+		get_uid(id_table[base.uid]);
+		get_guid(id_table[base.guid]);
 
-		switch(inode.inode_type) {
-			case SQUASHFS_FILE_TYPE: {
-				int frag_bytes = inode.fragment ==
-					SQUASHFS_INVALID_FRAG ? 0 :
-					inode.file_size % sBlk->block_size;
-				int blocks = inode.fragment ==
-					SQUASHFS_INVALID_FRAG ? (inode.file_size
-					+ sBlk->block_size - 1) >>
-					sBlk->block_log : inode.file_size >>
-					sBlk->block_log;
-				long long file_bytes = 0;
-				int i;
-				long long start = inode.start_block;
-				unsigned int *block_list;
+		switch(base.inode_type) {
+		case SQUASHFS_FILE_TYPE: {
+			struct squashfs_reg_inode_header inode;
+			int frag_bytes, blocks, i;
+			long long start, file_bytes = 0;
+			unsigned int *block_list;
 
-				TRACE("scan_inode_table: regular file, "
-					"file_size %d, blocks %d\n",
-					inode.file_size, blocks);
+			if(NO_INODE_BYTES(squashfs_reg_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
 
-				block_list = malloc(blocks *
-					sizeof(unsigned int));
-				if(block_list == NULL) {
-					ERROR("Out of memory in block list "
-						"malloc\n");
-					goto failed;
-				}
+			SQUASHFS_SWAP_REG_INODE_HEADER(cur_ptr, &inode);
 
-				cur_ptr += sizeof(inode);
-				SQUASHFS_SWAP_INTS(block_list, cur_ptr, blocks);
+			frag_bytes = inode.fragment == SQUASHFS_INVALID_FRAG ?
+				0 : inode.file_size % sBlk->block_size;
+			blocks = inode.fragment == SQUASHFS_INVALID_FRAG ?
+				(inode.file_size + sBlk->block_size - 1) >>
+				sBlk->block_log : inode.file_size >>
+				sBlk->block_log;
+			start = inode.start_block;
 
-				*uncompressed_file += inode.file_size;
-				(*file_count) ++;
+			TRACE("scan_inode_table: regular file, file_size %d, "
+				"blocks %d\n", inode.file_size, blocks);
 
-				for(i = 0; i < blocks; i++)
-					file_bytes +=
-						SQUASHFS_COMPRESSED_SIZE_BLOCK
-						(block_list[i]);
+			if(NO_BYTES(blocks * sizeof(unsigned int)))
+				/* corrupted filesystem */
+				goto corrupted;
 
-	                        add_file(start, inode.file_size, file_bytes,
-					block_list, blocks, inode.fragment,
-					inode.offset, frag_bytes);
-				cur_ptr += blocks * sizeof(unsigned int);
-				break;
-			}	
-			case SQUASHFS_LREG_TYPE: {
-				struct squashfs_lreg_inode_header inode;
-				int frag_bytes;
-				int blocks;
-				long long file_bytes = 0;
-				int i;
-				long long start;
-				unsigned int *block_list;
+			block_list = malloc(blocks * sizeof(unsigned int));
+			if(block_list == NULL)
+				MEM_ERROR();
 
-				SQUASHFS_SWAP_LREG_INODE_HEADER(&inode, cur_ptr);
+			cur_ptr += sizeof(inode);
+			SQUASHFS_SWAP_INTS(cur_ptr, block_list, blocks);
 
-				frag_bytes = inode.fragment ==
-					SQUASHFS_INVALID_FRAG ? 0 :
-					inode.file_size % sBlk->block_size;
-				blocks = inode.fragment == SQUASHFS_INVALID_FRAG
-					?  (inode.file_size +
-					sBlk->block_size - 1) >>
-					sBlk->block_log : inode.file_size >>
-					sBlk->block_log;
-				start = inode.start_block;
+			*uncompressed_file += inode.file_size;
+			(*file_count) ++;
 
-				TRACE("scan_inode_table: extended regular "
-					"file, file_size %lld, blocks %d\n",
-					inode.file_size, blocks);
+			for(i = 0; i < blocks; i++)
+				file_bytes +=
+					SQUASHFS_COMPRESSED_SIZE_BLOCK
+								(block_list[i]);
 
-				block_list = malloc(blocks *
-					sizeof(unsigned int));
-				if(block_list == NULL) {
-					ERROR("Out of memory in block list "
-						"malloc\n");
-					goto failed;
-				}
-
-				cur_ptr += sizeof(inode);
-				SQUASHFS_SWAP_INTS(block_list, cur_ptr, blocks);
-
-				*uncompressed_file += inode.file_size;
-				(*file_count) ++;
-
-				for(i = 0; i < blocks; i++)
-					file_bytes +=
-						SQUASHFS_COMPRESSED_SIZE_BLOCK
-						(block_list[i]);
-
-	                        add_file(start, inode.file_size, file_bytes,
-					block_list, blocks, inode.fragment,
-					inode.offset, frag_bytes);
-				cur_ptr += blocks * sizeof(unsigned int);
-				break;
-			}	
-			case SQUASHFS_SYMLINK_TYPE:
-			case SQUASHFS_LSYMLINK_TYPE: {
-				struct squashfs_symlink_inode_header inodep;
-	
-				SQUASHFS_SWAP_SYMLINK_INODE_HEADER(&inodep,
-					cur_ptr);
-				(*sym_count) ++;
-				cur_ptr += sizeof(inodep) + inodep.symlink_size;
-
-				if (inode.inode_type == SQUASHFS_LSYMLINK_TYPE)
-					cur_ptr += sizeof(unsigned int);
-
-				break;
+			if(inode.fragment != SQUASHFS_INVALID_FRAG &&
+					inode.fragment >= sBlk->fragments) {
+				free(block_list);
+				goto corrupted;
 			}
-			case SQUASHFS_DIR_TYPE: {
-				struct squashfs_dir_inode_header dir_inode;
 
-				SQUASHFS_SWAP_DIR_INODE_HEADER(&dir_inode,
-					cur_ptr);
-				if(dir_inode.start_block < directory_start_block)
-					*uncompressed_directory +=
-					dir_inode.file_size;
-				(*dir_count) ++;
-				cur_ptr += sizeof(struct squashfs_dir_inode_header);
-				break;
+			add_file(start, inode.file_size, file_bytes,
+				block_list, blocks, inode.fragment,
+				inode.offset, frag_bytes);
+				
+			cur_ptr += blocks * sizeof(unsigned int);
+			break;
+		}	
+		case SQUASHFS_LREG_TYPE: {
+			struct squashfs_lreg_inode_header inode;
+			int frag_bytes, blocks, i;
+			long long start, file_bytes = 0;
+			unsigned int *block_list;
+
+			if(NO_INODE_BYTES(squashfs_lreg_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			SQUASHFS_SWAP_LREG_INODE_HEADER(cur_ptr, &inode);
+
+			frag_bytes = inode.fragment == SQUASHFS_INVALID_FRAG ?
+				0 : inode.file_size % sBlk->block_size;
+			blocks = inode.fragment == SQUASHFS_INVALID_FRAG ?
+				(inode.file_size + sBlk->block_size - 1) >>
+				sBlk->block_log : inode.file_size >>
+				sBlk->block_log;
+			start = inode.start_block;
+
+			TRACE("scan_inode_table: extended regular "
+				"file, file_size %lld, blocks %d\n",
+				inode.file_size, blocks);
+
+			if(NO_BYTES(blocks * sizeof(unsigned int)))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			block_list = malloc(blocks * sizeof(unsigned int));
+			if(block_list == NULL)
+				MEM_ERROR();
+
+			cur_ptr += sizeof(inode);
+			SQUASHFS_SWAP_INTS(cur_ptr, block_list, blocks);
+
+			*uncompressed_file += inode.file_size;
+			(*file_count) ++;
+
+			for(i = 0; i < blocks; i++)
+				file_bytes +=
+					SQUASHFS_COMPRESSED_SIZE_BLOCK
+								(block_list[i]);
+
+			if(inode.fragment != SQUASHFS_INVALID_FRAG &&
+					inode.fragment >= sBlk->fragments) {
+				free(block_list);
+				goto corrupted;
 			}
-			case SQUASHFS_LDIR_TYPE: {
-				struct squashfs_ldir_inode_header dir_inode;
-				int i;
 
-				SQUASHFS_SWAP_LDIR_INODE_HEADER(&dir_inode,
-					cur_ptr);
-				if(dir_inode.start_block < directory_start_block)
-					*uncompressed_directory +=
-					dir_inode.file_size;
-				(*dir_count) ++;
-				cur_ptr += sizeof(struct squashfs_ldir_inode_header);
-				for(i = 0; i < dir_inode.i_count; i++) {
-					struct squashfs_dir_index index;
+			add_file(start, inode.file_size, file_bytes,
+				block_list, blocks, inode.fragment,
+				inode.offset, frag_bytes);
 
-					SQUASHFS_SWAP_DIR_INDEX(&index,
-						cur_ptr);
-					cur_ptr += sizeof(struct squashfs_dir_index) +
-						index.size + 1;
-				}
-				break;
+			cur_ptr += blocks * sizeof(unsigned int);
+			break;
+		}	
+		case SQUASHFS_SYMLINK_TYPE:
+		case SQUASHFS_LSYMLINK_TYPE: {
+			struct squashfs_symlink_inode_header inode;
+
+			if(NO_INODE_BYTES(squashfs_symlink_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			SQUASHFS_SWAP_SYMLINK_INODE_HEADER(cur_ptr, &inode);
+
+			(*sym_count) ++;
+
+			if (inode.inode_type == SQUASHFS_LSYMLINK_TYPE) {
+				if(NO_BYTES(inode.symlink_size +
+							sizeof(unsigned int)))
+					/* corrupted filesystem */
+					goto corrupted;
+				cur_ptr += sizeof(inode) + inode.symlink_size +
+							sizeof(unsigned int);
+			} else {
+				if(NO_BYTES(inode.symlink_size))
+					/* corrupted filesystem */
+					goto corrupted;
+				cur_ptr += sizeof(inode) + inode.symlink_size;
 			}
-		 	case SQUASHFS_BLKDEV_TYPE:
-		 	case SQUASHFS_CHRDEV_TYPE:
-				(*dev_count) ++;
-				cur_ptr += sizeof(struct squashfs_dev_inode_header);
-				break;
-		 	case SQUASHFS_LBLKDEV_TYPE:
-		 	case SQUASHFS_LCHRDEV_TYPE:
-				(*dev_count) ++;
-				cur_ptr += sizeof(struct squashfs_ldev_inode_header);
-				break;
-			case SQUASHFS_FIFO_TYPE:
-				(*fifo_count) ++;
-				cur_ptr += sizeof(struct squashfs_ipc_inode_header);
-				break;
-			case SQUASHFS_LFIFO_TYPE:
-				(*fifo_count) ++;
-				cur_ptr += sizeof(struct squashfs_lipc_inode_header);
-				break;
-			case SQUASHFS_SOCKET_TYPE:
-				(*sock_count) ++;
-				cur_ptr += sizeof(struct squashfs_ipc_inode_header);
-				break;
-			case SQUASHFS_LSOCKET_TYPE:
-				(*sock_count) ++;
-				cur_ptr += sizeof(struct squashfs_lipc_inode_header);
-				break;
-		 	default:
-				ERROR("Unknown inode type %d in "
-					"scan_inode_table!\n",
-					inode.inode_type);
-				goto failed;
+			break;
+		}
+		case SQUASHFS_DIR_TYPE: {
+			struct squashfs_dir_inode_header dir_inode;
+
+			if(NO_INODE_BYTES(squashfs_dir_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+				
+			SQUASHFS_SWAP_DIR_INODE_HEADER(cur_ptr, &dir_inode);
+
+			if(dir_inode.start_block < directory_start_block)
+				*uncompressed_directory += dir_inode.file_size;
+
+			(*dir_count) ++;
+			cur_ptr += sizeof(struct squashfs_dir_inode_header);
+			break;
+		}
+		case SQUASHFS_LDIR_TYPE: {
+			struct squashfs_ldir_inode_header dir_inode;
+			int i;
+
+			if(NO_INODE_BYTES(squashfs_ldir_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			SQUASHFS_SWAP_LDIR_INODE_HEADER(cur_ptr, &dir_inode);
+
+			if(dir_inode.start_block < directory_start_block)
+				*uncompressed_directory += dir_inode.file_size;
+
+			(*dir_count) ++;
+			cur_ptr += sizeof(struct squashfs_ldir_inode_header);
+
+			for(i = 0; i < dir_inode.i_count; i++) {
+				struct squashfs_dir_index index;
+
+				if(NO_BYTES(sizeof(index)))
+					/* corrupted filesystem */
+					goto corrupted;
+			
+				SQUASHFS_SWAP_DIR_INDEX(cur_ptr, &index);
+
+				if(NO_BYTES(index.size + 1))
+					/* corrupted filesystem */
+					goto corrupted;
+
+				cur_ptr += sizeof(index) + index.size + 1;
+			}
+			break;
+		}
+	 	case SQUASHFS_BLKDEV_TYPE:
+	 	case SQUASHFS_CHRDEV_TYPE:
+			if(NO_INODE_BYTES(squashfs_dev_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			(*dev_count) ++;
+			cur_ptr += sizeof(struct squashfs_dev_inode_header);
+			break;
+	 	case SQUASHFS_LBLKDEV_TYPE:
+	 	case SQUASHFS_LCHRDEV_TYPE:
+			if(NO_INODE_BYTES(squashfs_ldev_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			(*dev_count) ++;
+			cur_ptr += sizeof(struct squashfs_ldev_inode_header);
+			break;
+		case SQUASHFS_FIFO_TYPE:
+			if(NO_INODE_BYTES(squashfs_ipc_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			(*fifo_count) ++;
+			cur_ptr += sizeof(struct squashfs_ipc_inode_header);
+			break;
+		case SQUASHFS_LFIFO_TYPE:
+			if(NO_INODE_BYTES(squashfs_lipc_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			(*fifo_count) ++;
+			cur_ptr += sizeof(struct squashfs_lipc_inode_header);
+			break;
+		case SQUASHFS_SOCKET_TYPE:
+			if(NO_INODE_BYTES(squashfs_ipc_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			(*sock_count) ++;
+			cur_ptr += sizeof(struct squashfs_ipc_inode_header);
+			break;
+		case SQUASHFS_LSOCKET_TYPE:
+			if(NO_INODE_BYTES(squashfs_lipc_inode_header))
+				/* corrupted filesystem */
+				goto corrupted;
+
+			(*sock_count) ++;
+			cur_ptr += sizeof(struct squashfs_lipc_inode_header);
+			break;
+	 	default:
+			ERROR("Unknown inode type %d in scan_inode_table!\n",
+					base.inode_type);
+			goto corrupted;
 		}
 	}
 	
 	printf("Read existing filesystem, %d inodes scanned\n", files);
 	return TRUE;
 
-
-failed:
+corrupted:
+	ERROR("scan_inode_table: filesystem corruption detected in "
+		"scanning metadata\n");
 	free(*inode_table);
 	return FALSE;
 }
@@ -379,8 +500,12 @@ struct compressor *read_super(int fd, struct squashfs_super_block *sBlk, char *s
 
 	res = read_fs_bytes(fd, SQUASHFS_START, sizeof(struct squashfs_super_block),
 		sBlk);
-	if(res == 0)
+	if(res == 0) {
+		ERROR("Can't find a SQUASHFS superblock on %s\n",
+				source);
+		ERROR("Wrong filesystem or filesystem is corrupted!\n");
 		goto failed_mount;
+	}
 
 	SQUASHFS_INSWAP_SUPER_BLOCK(sBlk);
 
@@ -388,9 +513,11 @@ struct compressor *read_super(int fd, struct squashfs_super_block *sBlk, char *s
 		if(sBlk->s_magic == SQUASHFS_MAGIC_SWAP)
 			ERROR("Pre 4.0 big-endian filesystem on %s, appending"
 				" to this is unsupported\n", source);
-		else
+		else {
 			ERROR("Can't find a SQUASHFS superblock on %s\n",
 				source);
+			ERROR("Wrong filesystem or filesystem is corrupted!\n");
+		}
 		goto failed_mount;
 	}
 
@@ -413,7 +540,7 @@ struct compressor *read_super(int fd, struct squashfs_super_block *sBlk, char *s
 	/* Check the compression type */
 	comp = lookup_compressor_id(sBlk->compression);
 	if(!comp->supported) {
-		ERROR("Filesystem on %s uses %s compression, this is"
+		ERROR("Filesystem on %s uses %s compression, this is "
 			"unsupported by this version\n", source, comp->name);
 		ERROR("Compressors available:\n");
 		display_compressors("", "");
@@ -430,12 +557,20 @@ struct compressor *read_super(int fd, struct squashfs_super_block *sBlk, char *s
 	 * is still called to set the default options (the defaults may have
 	 * been changed by the user specifying options on the command
 	 * line which need to be over-ridden).
+	 *
+	 * Compressor_extract_options is also used to ensure that 
+	 * we know how decompress a filesystem compressed with these
+	 * compression options.
 	 */
 	if(SQUASHFS_COMP_OPTS(sBlk->flags)) {
-		bytes = read_block(fd, sizeof(*sBlk), NULL, buffer);
+		bytes = read_block(fd, sizeof(*sBlk), NULL, 0, buffer);
 
-		if(bytes == 0)
+		if(bytes == 0) {
+			ERROR("Failed to read compressor options from append "
+				"filesystem\n");
+			ERROR("Filesystem corrupted?\n");
 			goto failed_mount;
+		}
 	}
 
 	res = compressor_extract_options(comp, sBlk->block_size, buffer, bytes);
@@ -457,7 +592,7 @@ struct compressor *read_super(int fd, struct squashfs_super_block *sBlk, char *s
 		SQUASHFS_UNCOMPRESSED_XATTRS(sBlk->flags) ? "un" : "");
 	printf("\tFragments are %spresent in the filesystem\n",
 		SQUASHFS_NO_FRAGMENTS(sBlk->flags) ? "not " : "");
-	printf("\tAlways_use_fragments option is %sspecified\n",
+	printf("\tAlways-use-fragments option is %sspecified\n",
 		SQUASHFS_ALWAYS_FRAGMENTS(sBlk->flags) ? "" : "not ");
 	printf("\tDuplicates are %sremoved\n",
 		SQUASHFS_DUPLICATES(sBlk->flags) ? "" : "not ");
@@ -504,13 +639,20 @@ unsigned char *squashfs_readdir(int fd, int root_entries,
 	directory_table = malloc((size + SQUASHFS_METADATA_SIZE * 2 - 1) &
 		~(SQUASHFS_METADATA_SIZE - 1));
 	if(directory_table == NULL)
-		return NULL;
+		MEM_ERROR();
+
 	while(bytes < size) {
+		int expected = (size - bytes) >= SQUASHFS_METADATA_SIZE ?
+			SQUASHFS_METADATA_SIZE : 0;
+
 		TRACE("squashfs_readdir: reading block 0x%llx, bytes read so "
 			"far %d\n", start, bytes);
+
 		last_start_block = start;
-		byte = read_block(fd, start, &start, directory_table + bytes);
+		byte = read_block(fd, start, &start, expected, directory_table + bytes);
 		if(byte == 0) {
+			ERROR("Failed to read directory\n");
+			ERROR("Filesystem corrupted?\n");
 			free(directory_table);
 			return NULL;
 		}
@@ -522,7 +664,7 @@ unsigned char *squashfs_readdir(int fd, int root_entries,
 
 	bytes = offset;
  	while(bytes < size) {			
-		SQUASHFS_SWAP_DIR_HEADER(&dirh, directory_table + bytes);
+		SQUASHFS_SWAP_DIR_HEADER(directory_table + bytes, &dirh);
 
 		dir_count = dirh.count + 1;
 		TRACE("squashfs_readdir: Read directory header @ byte position "
@@ -530,7 +672,7 @@ unsigned char *squashfs_readdir(int fd, int root_entries,
 		bytes += sizeof(dirh);
 
 		while(dir_count--) {
-			SQUASHFS_SWAP_DIR_ENTRY(dire, directory_table + bytes);
+			SQUASHFS_SWAP_DIR_ENTRY(directory_table + bytes, dire);
 			bytes += sizeof(*dire);
 
 			memcpy(dire->name, directory_table + bytes,
@@ -563,14 +705,14 @@ unsigned int *read_id_table(int fd, struct squashfs_super_block *sBlk)
 	int res, i;
 
 	id_table = malloc(bytes);
-	if(id_table == NULL) {
-		ERROR("Failed to allocate id table\n");
-		return NULL;
-	}
+	if(id_table == NULL)
+		MEM_ERROR();
 
 	res = read_fs_bytes(fd, sBlk->id_table_start,
 		SQUASHFS_ID_BLOCK_BYTES(sBlk->no_ids), index);
 	if(res == 0) {
+		ERROR("Failed to read id table index\n");
+		ERROR("Filesystem corrupted?\n");
 		free(id_table);
 		return NULL;
 	}
@@ -578,7 +720,9 @@ unsigned int *read_id_table(int fd, struct squashfs_super_block *sBlk)
 	SQUASHFS_INSWAP_ID_BLOCKS(index, indexes);
 
 	for(i = 0; i < indexes; i++) {
-		int length = read_block(fd, index[i], NULL,
+		int expected = (i + 1) != indexes ? SQUASHFS_METADATA_SIZE :
+					bytes & (SQUASHFS_METADATA_SIZE - 1);
+		int length = read_block(fd, index[i], NULL, expected,
 			((unsigned char *) id_table) +
 			(i * SQUASHFS_METADATA_SIZE));
 		TRACE("Read id table block %d, from 0x%llx, length %d\n", i,
@@ -586,6 +730,7 @@ unsigned int *read_id_table(int fd, struct squashfs_super_block *sBlk)
 		if(length == 0) {
 			ERROR("Failed to read id table block %d, from 0x%llx, "
 				"length %d\n", i, index[i], length);
+			ERROR("Filesystem corrupted?\n");
 			free(id_table);
 			return NULL;
 		}
@@ -605,26 +750,28 @@ unsigned int *read_id_table(int fd, struct squashfs_super_block *sBlk)
 int read_fragment_table(int fd, struct squashfs_super_block *sBlk,
 	struct squashfs_fragment_entry **fragment_table)
 {
-	int res, i, indexes = SQUASHFS_FRAGMENT_INDEXES(sBlk->fragments);
+	int res, i;
+	int bytes = SQUASHFS_FRAGMENT_BYTES(sBlk->fragments);
+	int indexes = SQUASHFS_FRAGMENT_INDEXES(sBlk->fragments);
 	long long fragment_table_index[indexes];
 
 	TRACE("read_fragment_table: %d fragments, reading %d fragment indexes "
 		"from 0x%llx\n", sBlk->fragments, indexes,
 		sBlk->fragment_table_start);
+
 	if(sBlk->fragments == 0)
 		return 1;
 
-	*fragment_table = malloc(sBlk->fragments *
-		sizeof(struct squashfs_fragment_entry));
-	if(*fragment_table == NULL) {
-		ERROR("Failed to allocate fragment table\n");
-		return 0;
-	}
+	*fragment_table = malloc(bytes);
+	if(*fragment_table == NULL)
+		MEM_ERROR();
 
 	res = read_fs_bytes(fd, sBlk->fragment_table_start,
 		SQUASHFS_FRAGMENT_INDEX_BYTES(sBlk->fragments),
 		fragment_table_index);
 	if(res == 0) {
+		ERROR("Failed to read fragment table index\n");
+		ERROR("Filesystem corrupted?\n");
 		free(*fragment_table);
 		return 0;
 	}
@@ -632,8 +779,10 @@ int read_fragment_table(int fd, struct squashfs_super_block *sBlk,
 	SQUASHFS_INSWAP_FRAGMENT_INDEXES(fragment_table_index, indexes);
 
 	for(i = 0; i < indexes; i++) {
+		int expected = (i + 1) != indexes ? SQUASHFS_METADATA_SIZE :
+					bytes & (SQUASHFS_METADATA_SIZE - 1);
 		int length = read_block(fd, fragment_table_index[i], NULL,
-			((unsigned char *) *fragment_table) +
+			expected, ((unsigned char *) *fragment_table) +
 			(i * SQUASHFS_METADATA_SIZE));
 		TRACE("Read fragment table block %d, from 0x%llx, length %d\n",
 			i, fragment_table_index[i], length);
@@ -641,6 +790,7 @@ int read_fragment_table(int fd, struct squashfs_super_block *sBlk,
 			ERROR("Failed to read fragment table block %d, from "
 				"0x%llx, length %d\n", i,
 				fragment_table_index[i], length);
+			ERROR("Filesystem corrupted?\n");
 			free(*fragment_table);
 			return 0;
 		}
@@ -665,14 +815,14 @@ int read_inode_lookup_table(int fd, struct squashfs_super_block *sBlk,
 		return 1;
 
 	*inode_lookup_table = malloc(lookup_bytes);
-	if(*inode_lookup_table == NULL) {
-		ERROR("Failed to allocate inode lookup table\n");
-		return 0;
-	}
+	if(*inode_lookup_table == NULL)
+		MEM_ERROR();
 
 	res = read_fs_bytes(fd, sBlk->lookup_table_start,
 		SQUASHFS_LOOKUP_BLOCK_BYTES(sBlk->inodes), index);
 	if(res == 0) {
+		ERROR("Failed to read inode lookup table index\n");
+		ERROR("Filesystem corrupted?\n");
 		free(*inode_lookup_table);
 		return 0;
 	}
@@ -680,7 +830,9 @@ int read_inode_lookup_table(int fd, struct squashfs_super_block *sBlk,
 	SQUASHFS_INSWAP_LONG_LONGS(index, indexes);
 
 	for(i = 0; i <  indexes; i++) {
-		int length = read_block(fd, index[i], NULL,
+		int expected = (i + 1) != indexes ? SQUASHFS_METADATA_SIZE :
+				lookup_bytes & (SQUASHFS_METADATA_SIZE - 1);
+		int length = read_block(fd, index[i], NULL, expected,
 			((unsigned char *) *inode_lookup_table) +
 			(i * SQUASHFS_METADATA_SIZE));
 		TRACE("Read inode lookup table block %d, from 0x%llx, length "
@@ -689,6 +841,7 @@ int read_inode_lookup_table(int fd, struct squashfs_super_block *sBlk,
 			ERROR("Failed to read inode lookup table block %d, "
 				"from 0x%llx, length %d\n", i, index[i],
 				length);
+			ERROR("Filesystem corrupted?\n");
 			free(*inode_lookup_table);
 			return 0;
 		}
@@ -714,7 +867,7 @@ long long read_filesystem(char *root_name, int fd, struct squashfs_super_block *
 	struct squashfs_fragment_entry **fragment_table,
 	squashfs_inode **inode_lookup_table)
 {
-	unsigned char *inode_table = NULL, *directory_table;
+	unsigned char *inode_table = NULL, *directory_table = NULL;
 	long long start = sBlk->inode_table_start;
 	long long end = sBlk->directory_table_start;
 	long long root_inode_start = start +
@@ -723,7 +876,7 @@ long long read_filesystem(char *root_name, int fd, struct squashfs_super_block *
 		SQUASHFS_INODE_OFFSET(sBlk->root_inode);
 	unsigned int root_inode_block;
 	union squashfs_inode_header inode;
-	unsigned int *id_table;
+	unsigned int *id_table = NULL;
 	int res;
 
 	printf("Scanning existing filesystem...\n");
@@ -746,10 +899,8 @@ long long read_filesystem(char *root_name, int fd, struct squashfs_super_block *
 		&root_inode_block, root_inode_size, uncompressed_file,
 		uncompressed_directory, file_count, sym_count, dev_count,
 		dir_count, fifo_count, sock_count, id_table);
-	if(res == 0) {
-		ERROR("read_filesystem: inode table read failed\n");
+	if(res == 0)
 		goto error;
-	}
 
 	*uncompressed_inode = root_inode_block;
 
@@ -773,57 +924,57 @@ long long read_filesystem(char *root_name, int fd, struct squashfs_super_block *
 			*inode_dir_start_block, *inode_dir_offset,
 			*inode_dir_file_size, last_directory_block, sBlk,
 			push_directory_entry);
-		if(directory_table == NULL) {
-			ERROR("read_filesystem: Could not read root directory"
-				"\n");
+		if(directory_table == NULL) 
 			goto error;
-		}
 
 		root_inode_start -= start;
 		*cinode_table = malloc(root_inode_start);
-		if(*cinode_table == NULL) {
-			ERROR("read_filesystem: failed to alloc space for "
-				"existing filesystem inode table\n");
+		if(*cinode_table == NULL)
+			MEM_ERROR();
+
+	       	res = read_fs_bytes(fd, start, root_inode_start, *cinode_table);
+		if(res == 0) {
+			ERROR("Failed to read inode table\n");
+			ERROR("Filesystem corrupted?\n");
 			goto error;
 		}
-	       	res = read_fs_bytes(fd, start, root_inode_start, *cinode_table);
-		if(res == 0)
-			goto error;
 
 		*cdirectory_table = malloc(*last_directory_block);
-		if(*cdirectory_table == NULL) {
-			ERROR("read_filesystem: failed to alloc space for "
-				"existing filesystem directory table\n");
-			goto error;
-		}
+		if(*cdirectory_table == NULL)
+			MEM_ERROR();
+
 		res = read_fs_bytes(fd, sBlk->directory_table_start,
 			*last_directory_block, *cdirectory_table);
-		if(res == 0)
-			goto error;
-
-		*data_cache = malloc(root_inode_offset + *root_inode_size);
-		if(*data_cache == NULL) {
-			ERROR("read_filesystem: failed to alloc inode cache\n");
+		if(res == 0) {
+			ERROR("Failed to read directory table\n");
+			ERROR("Filesystem corrupted?\n");
 			goto error;
 		}
+
+		*data_cache = malloc(root_inode_offset + *root_inode_size);
+		if(*data_cache == NULL)
+			MEM_ERROR();
+
 		memcpy(*data_cache, inode_table + root_inode_block,
 			root_inode_offset + *root_inode_size);
 
 		*directory_data_cache = malloc(*inode_dir_offset +
 			*inode_dir_file_size);
-		if(*directory_data_cache == NULL) {
-			ERROR("read_filesystem: failed to alloc directory "
-				"cache\n");
-			goto error;
-		}
+		if(*directory_data_cache == NULL)
+			MEM_ERROR();
+
 		memcpy(*directory_data_cache, directory_table,
 			*inode_dir_offset + *inode_dir_file_size);
 
+		free(id_table);
 		free(inode_table);
 		free(directory_table);
 		return sBlk->inode_table_start;
 	}
 
 error:
+	free(id_table);
+	free(inode_table);
+	free(directory_table);
 	return 0;
 }
