@@ -3,8 +3,8 @@
 /*
  * Squashfs
  *
- * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010
- * Phillip Lougher <phillip@lougher.demon.co.uk>
+ * Copyright (c) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2013
+ * Phillip Lougher <phillip@squashfs.org.uk>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -23,20 +23,12 @@
  * read_fs.h
  *
  */
-
-#if __BYTE_ORDER == __BIG_ENDIAN
-#define SQUASHFS_SWAP_SHORTS(d, s, n) swap_le16_num(s, d, n)
-#define SQUASHFS_SWAP_INTS(d, s, n) swap_le32_num(s, d, n)
-#define SQUASHFS_SWAP_LONG_LONGS(d, s, n) swap_le64_num(s, d, n)
-
-#define SWAP_LE16(d, s)		swap_le16(s, d)
-#define SWAP_LE32(d, s)		swap_le32(s, d)
-#define SWAP_LE64(d, s)		swap_le64(s, d)
-#else
-#define SQUASHFS_MEMCPY(d, s, n)	memcpy(d, s, n)
-#define SQUASHFS_SWAP_SHORTS(d, s, n)	memcpy(d, s, n * sizeof(short))
-#define SQUASHFS_SWAP_INTS(d, s, n)	memcpy(d, s, n * sizeof(int))
-#define SQUASHFS_SWAP_LONG_LONGS(d, s, n) \
-					memcpy(d, s, n * sizeof(long long))
-#endif
+extern struct compressor *read_super(int, struct squashfs_super_block *,
+	char *);
+extern long long read_filesystem(char *, int, struct squashfs_super_block *,
+char **, char **, char **, char **, unsigned int *, unsigned int *,
+unsigned int *, unsigned int *, unsigned int *, int *, int *, int *, int *,
+int *, int *, long long *, unsigned int *, unsigned int *, unsigned int *,
+unsigned int *, void (push_directory_entry)(char *, squashfs_inode, int, int),
+struct squashfs_fragment_entry **, squashfs_inode **);
 #endif
